"""
Main master handler for each user of Jaseci, serves as main interface between
between user and Jaseci
"""

from jaseci.prim.element import Element
from jaseci.extens.api.alias_api import AliasAPI
from jaseci.extens.api.object_api import ObjectApi
from jaseci.extens.api.graph_api import GraphApi
from jaseci.extens.api.sentinel_api import SentinelApi
from jaseci.extens.api.walker_api import WalkerApi
from jaseci.extens.api.architype_api import ArchitypeApi
from jaseci.extens.api.config_api import ConfigApi
from jaseci.extens.api.interface import Interface
from jaseci.extens.api.master_api import MasterApi
from jaseci.extens.api.jac_api import JacApi
from jaseci.extens.api.user_api import UserApi
from jaseci.extens.api.queue_api import QueueApi
from jaseci.extens.api.webhook_api import WebhookApi
from jaseci.jsorc.jsorc import JsOrc


@JsOrc.context(name="master")
class Master(
    Element,
    Interface,
    MasterApi,
    AliasAPI,
    GraphApi,
    ObjectApi,
    SentinelApi,
    WalkerApi,
    ArchitypeApi,
    JacApi,
    UserApi,
    QueueApi,
    WebhookApi,
):
    """Main class for master functions for user"""

    def __init__(self, head_master=None, *args, **kwargs):
        kwargs["m_id"] = None

        Element.__init__(self, kind="Jaseci Master", *args, **kwargs)
        MasterApi.__init__(self, head_master)
        AliasAPI.__init__(self)
        ConfigApi.__init__(self)
        ObjectApi.__init__(self)
        GraphApi.__init__(self)
        WalkerApi.__init__(self)
        SentinelApi.__init__(self)
        Interface.__init__(self)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        GraphApi.destroy(self)
        SentinelApi.destroy(self)
        MasterApi.destroy(self)
        WalkerApi.destroy(self)
        super().destroy()



from jaseci.utils.test_core import CoreTest


class Lang14Test(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_free_reference(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("fam.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "free_ref"}])
        self.assertTrue(ret["success"])
        self.assertEqual(len(ret["report"]), 4)
        self.assertEqual(ret["report"][0][0]["name"], "woman")
        self.assertNotEqual(ret["report"][1][0], ret["report"][2][0])
        self.assertEqual(len(ret["report"][1]), 1)
        self.assertEqual(len(ret["report"][3]), 3)
        self.assertIn(ret["report"][1][0], ret["report"][3])

    def test_kwargs(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("general.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "test_kwargs"}])
        self.assertGreater(ret["report"][0].count("."), 2)
        self.assertLess(ret["report"][0].count("."), 6)
        self.assertEqual(ret["report"][1], ret["report"][2])
        self.assertNotEqual(ret["report"][3], ret["report"][2])

    def test_multistring(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("general.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "multistring"}])
        self.assertEqual(ret["report"][0], "This is a multistring")

    def test_free_ref_ifs(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("free_refs.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "init"}])
        self.assertEqual(ret["report"], [True, True])
        self.assertEqual(ret["success"], True)



from jaseci.jac.jsci_vm.op_codes import JsOp
from struct import unpack


def from_bytes(typ, val):
    if typ == str:
        if val is None:
            return ""
        return val.decode("unicode_escape")
    if typ == float:
        return unpack("d", val)[0]
    else:
        if val is None:
            return 0
        return typ.from_bytes(val, "little")


class InstPtr:
    def __init__(self):
        self._ip = 0
        self._bytecode = None

    def offset(self, delta, range=None):
        if range == 0:
            return None
        elif range is not None:
            return self._bytecode[self._ip + delta : self._ip + delta + range]
        return self._bytecode[self._ip + delta]

    def cur_op(self):
        return JsOp(self.offset(0)).name



from jaseci.utils.test_core import CoreTest
import os


class FileLibTest(CoreTest):
    fixture_src = __file__

    def test_json_dump(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("file_stuff.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "pack_it"}])
        self.assertEqual(ret["report"][0], {"hello": 5})

    def test_bin_load_save(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("file_stuff.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "bin_load_save"}])
        self.assertEqual(
            ret["report"][0],
            "iVBORw0KGgoAAAANSUhEUgAAAAgAAAAIAQMAAAD+wSzIAAAABlBMVEX"
            "///+/v7+jQ3Y5AAAADklEQVQI12P4AIX8EAgALgAD/aNpbtEAAAAASUVORK5CYII=",
        )
        self.assertFalse(os.path.exists("file.gif"))

    def test_kwargs_bin_load_save(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("file_stuff.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "kwargs_bin_load_save"}])
        self.assertEqual(
            ret["report"][0],
            "iVBORw0KGgoAAAANSUhEUgAAAAgAAAAIAQMAAAD+wSzIAAAABlBMVEX"
            "///+/v7+jQ3Y5AAAADklEQVQI12P4AIX8EAgALgAD/aNpbtEAAAAASUVORK5CYII=",
        )
        self.assertFalse(os.path.exists("file.gif"))



import os
import pytest
from unittest import TestCase

import jaseci.tests.jac_test_progs as jtp
from jaseci.prim.sentinel import Sentinel
from jaseci.prim.graph import Graph
from jaseci.prim.node import Node
from jaseci.utils.utils import TestCaseHelper
from jaseci.jsorc.jsorc import JsOrc
from jaseci.jsorc.jsorc_utils import State


class JacTests(TestCaseHelper, TestCase):
    """Unit tests for Jac language"""

    def setUp(self):
        super().setUp()

    def tearDown(self):
        super().tearDown()

    def test_bug_check1(self):
        self.logger_on()
        sent = Sentinel(m_id=0, h=JsOrc.hook())
        gph = Graph(m_id=0, h=sent._h)
        sent.register_code(jtp.bug_check1)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report[0][0], "THIS IS AN INTENT_LABEL")

    def test_action_load_std_lib(self):
        mast = JsOrc.super_master()
        mast.sentinel_register(name="test", code=jtp.action_load_std_lib)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "aload"}
        )["report"]
        self.assertEqual(report[0], True)

    def test_action_load_std_lib_only_super(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.action_load_std_lib)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "aload"}
        )
        report = report["report"]
        self.assertEqual(report[0], False)

    def test_globals(self):
        sent = Sentinel(m_id=0, h=JsOrc.hook())
        gph = Graph(m_id=0, h=sent._h)
        sent.register_code(jtp.globals)
        test_walker = sent.run_architype("init")
        test_walker.prime(gph)
        test_walker.run()
        report = test_walker.report
        self.assertEqual(report, ["testing", 56])

    def test_net_root_std_lib(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.net_root_std_lib)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(report[0][0], report[0][1])
        self.assertEqual(report[0][1], report[1][1])
        self.assertNotEqual(report[1][0], report[1][1])

    def test_or_stmt(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.or_stmt)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(report, [[3.4, "Hello"]])

    def test_nd_equals(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.nd_equals_error_correct_line)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertIn("line 3", report["errors"][0])

    def test_strange_ability_bug(self):
        import json

        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.strange_ability_bug)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "travel"}
        )["report"]
        mast.sentinel_register(name="test", code=jtp.strange_ability_bug, auto_run="")
        report += mast.general_interface_to_api(
            api_name="walker_run", params={"name": "travel"}
        )["report"]
        ir = mast.sentinel_get(mode="ir", snt=mast.active_snt())
        mast.sentinel_set(code=json.dumps(ir), snt=mast.active_snt(), mode="ir")
        report += mast.general_interface_to_api(
            api_name="walker_run", params={"name": "travel"}
        )["report"]
        report += mast.general_interface_to_api(
            api_name="walker_run", params={"name": "travel"}
        )["report"]
        self.assertEqual(report, ["Showing", "Showing", "Showing", "Showing"])

    def test_node_inheritance(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.node_inheritance, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        del report["final_node"]
        self.assertEqual(
            report,
            {
                "report": [
                    "plain.x",
                    {"a": 55, "b": 7, "c": 7, "d": 80},
                    "super.x",
                    "plain2.y",
                    "super.y",
                ],
                "success": True,
                "yielded": False,
            },
        )

    def test_inherited_ref(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.inherited_ref, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(len(report["report"]), 12)

    def test_node_inheritance_chain_check(self):
        mast = JsOrc.master()
        mast.sentinel_register(
            name="test", code=jtp.node_inheritance_chain_check, auto_run=""
        )
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(report["success"], False)

    def test_global_reregistering(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.global_reregistering)
        self.assertTrue(mast.active_snt().is_active)
        mast.sentinel_set(snt=mast.active_snt(), code=jtp.global_reregistering)
        self.assertTrue(mast.active_snt().is_active)
        mast.sentinel_register(name="test", code=jtp.global_reregistering)
        self.assertTrue(mast.active_snt().is_active)

    def test_vector_cos_sim_check(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.vector_cos_sim_check, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(len(report), 1)
        self.assertEqual(type(report[0]), float)

    def test_multi_breaks(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.multi_breaks, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(len(report), 15)
        self.assertEqual(report[14], 180)

    def test_reffy_deref_check(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.reffy_deref_check, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertFalse(report[0])
        self.assertTrue(report[1])

    def test_vanishing_can_check(self):
        mast = JsOrc.super_master()
        mast.actions_load_local("jaseci/tests/infer.py")
        mast.sentinel_register(name="test", code=jtp.vanishing_can_check, auto_run="")
        mast.general_interface_to_api(api_name="walker_run", params={"name": "init"})
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(report, ["2022-01-01T00:00:00"])
        mast.propagate_access("public")
        mast2 = JsOrc.super_master(h=mast._h)
        mast.active_snt().run_architype(name="plain", kind="node", caller=mast2)
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(report, ["2022-01-01T00:00:00"])

    def test_jasecilib_alias_list(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.jasecilib_alias_list, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertGreater(len(report[0].keys()), 3)

    def test_jasecilib_params(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.jasecilib_params, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertIn("j_r_acc_ids", report[0][0].keys())

    def test_jasecilib_create_user(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.jasecilib_create_user, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )["report"]
        self.assertEqual(report[0]["user"]["name"], "daman@gmail.com")

    def test_root_is_node_type(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.root_is_node_type, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(report["report"][0], "JAC_TYPE.NODE")

    def test_walker_with_exit_after_node(self):
        mast = JsOrc.master()
        mast.sentinel_register(
            name="test", code=jtp.walker_with_exit_after_node, auto_run=""
        )
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(report["report"], [1, 1, 3, 1, 3, 1, 3, 1, 3, 43])

    def test_depth_first_take(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.depth_first_take, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(report["report"], [1, 2, 3, 4, 5, 6, 7])

    def test_breadth_first_take(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.breadth_first_take, auto_run="")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(report["report"], [1, 2, 5, 3, 4, 6, 7])

    def test_inheritance_override_here_check(self):
        mast = JsOrc.master()
        mast.sentinel_register(
            name="test", code=jtp.inheritance_override_here_check, auto_run=""
        )
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(report["report"], [9, 9, 10])

    def test_dot_private_hidden(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.dot_private_hidden, auto_run="")
        mast.general_interface_to_api(api_name="walker_run", params={"name": "init"})
        report = mast.general_interface_to_api(
            api_name="graph_get", params={"mode": "dot", "detailed": True}
        )
        self.assertNotIn("j=", report)

    def test_check_destroy_node_has_var(self):
        mast = JsOrc.master()
        mast.sentinel_register(
            name="test", code=jtp.check_destroy_node_has_var, auto_run=""
        )
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "create"}
        )
        self.assertEqual(mast._h.get_object_distribution()[Node], 2)
        self.assertEqual(report["report"][0], "JAC_TYPE.NODE")
        report = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "remove"}
        )
        self.assertEqual(mast._h.get_object_distribution()[Node], 1)
        self.assertEqual(report["report"][0], "JAC_TYPE.NULL")

    def test_for_loop_dict(self):
        mast = JsOrc.master()
        mast.sentinel_register(
            name="test", code=jtp.check_dict_for_in_loop, auto_run=""
        )
        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "for_loop_dict"}
        )

        self.assertEqual(
            res["report"],
            [
                "test1 : 1",
                "test2 : 2",
                "test3 : 3",
                "test1 : 1",
                "test2 : 2",
                "test3 : 3",
                5,
                6,
                7,
                "0 : 5",
                "1 : 6",
                "2 : 7",
            ],
        )

    def test_var_as_key_for_dict(self):
        mast = JsOrc.master()
        mast.sentinel_register(
            name="test", code=jtp.check_dict_for_in_loop, auto_run=""
        )
        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "var_as_key_for_dict"}
        )

        self.assertEqual(res["report"], [{"key1": "key1", "key2": 2}])
        self.assertIn("Key is not str type : <class 'int'>!", res["errors"][0])

    def test_list_pairwise(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.list_pairwise, auto_run="")

        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )

        self.assertEqual(
            res["report"],
            [[[1, 2], [2, 3], [3, 4]]],
        )

    def test_list_unique(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.list_unique, auto_run="")

        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )

        self.assertEqual(
            res["report"],
            [[1, 2, 3, 4, 5]],
        )

    def test_new_additional_builtin(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.check_new_builtin, auto_run="")

        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )

        self.assertEqual(
            res["report"],
            [{"test": "test"}, 1, "1 2 3 4", "1 2 3 4"],
        )

    def test_continue_issue(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.continue_issue, auto_run="")
        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(res["report"], [1, 2, 3, 4, 5, 6, 7, 8, "apple"])

    def test_registering_dict_as_ir(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.continue_issue, auto_run="")
        code_dict = mast.sentinel_get(snt=mast.active_snt(), mode="ir")
        self.assertEqual(type(code_dict), dict)
        mast.sentinel_register(name="test", code=code_dict, mode="ir", auto_run="")
        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(res["report"], [1, 2, 3, 4, 5, 6, 7, 8, "apple"])

    @pytest.mark.order(2)
    def test_async_syntax_with_celery(self):
        mast = JsOrc.master()
        if not JsOrc.svc("task").is_running():
            self.skip_test("Celery not running")
        mast.sentinel_register(name="test", code=jtp.async_syntax, auto_run="")
        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "simple_async", "ctx": {"sample": "name"}},
        )

        self.assertTrue(res["is_queued"])

        res = mast.general_interface_to_api(
            api_name="walker_queue_wait",
            params={"task_id": res["result"], "timeout": 15},
        )

        self.assertEqual("test", res["result"]["anchor"])

        res = res["result"]["response"]

        # report from task1 (awaited)
        self.assertEqual(1, res["report"][0])
        self.assertEqual(2, res["report"][1])

        # task1 anchor
        self.assertEqual(1, res["report"][2])
        report4 = res["report"][4]

        res = res["report"][3]

        # no celery running
        self.assertTrue(res["is_queued"])

        res = mast.general_interface_to_api(
            api_name="walker_queue_check",
            params={"task_id": res["result"]},
        )

        self.assertEqual(report4, res)
        self.assertEqual(2, res["result"]["anchor"])
        self.assertEqual([2, 2], res["result"]["response"]["report"])

    @pytest.mark.order(3)
    def test_async_syntax_with_update_and_celery(self):
        mast = JsOrc.master()
        if not JsOrc.svc("task").is_running():
            self.skip_test("Celery not running")
        mast.sentinel_register(
            name="test", code=jtp.async_syntax_with_update, auto_run="init"
        )
        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "update_value", "ctx": {}},
        )

        self.assertTrue(res["is_queued"])

        res = mast.general_interface_to_api(
            api_name="walker_queue_wait",
            params={"task_id": res["result"], "timeout": 15},
        )

        self.assertEqual("SUCCESS", res["status"])
        self.assertTrue(res["result"]["response"]["report"][0]["value"])

        # remove the node on current _h.mem and let the redis repopulate it
        mast._h.mem.pop(res["result"]["response"]["final_node"], None)

        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "get_value", "ctx": {}},
        )

        self.assertTrue(res["report"][0]["value"])

    def test_async_syntax_without_celery(self):
        mast = JsOrc.master()
        JsOrc.svc("task").state = State.NOT_STARTED
        mast.sentinel_register(name="test", code=jtp.async_syntax, auto_run="")
        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "simple_async", "ctx": {"sample": "name"}},
        )

        # no celery running
        self.assertFalse(res["is_queued"])

        # report from task1 (awaited)
        self.assertEqual(1, res["result"]["report"][0])
        self.assertEqual(2, res["result"]["report"][1])

        # report from task2 (async but no celery)
        self.assertEqual(2, res["result"]["report"][2])
        self.assertEqual(2, res["result"]["report"][3])

        # task1 anchor
        self.assertEqual(1, res["result"]["report"][4])

        res = res["result"]["report"][5]

        # no celery running
        self.assertFalse(res["is_queued"])
        self.assertEqual(2, res["result"])

        JsOrc.svc("task").state = State.RUNNING

    def test_async_sync_syntax_with_celery(self):
        mast = JsOrc.master()
        if not JsOrc.svc("task").is_running():
            self.skip_test("Celery not running")
        mast.sentinel_register(name="test", code=jtp.async_syntax, auto_run="")
        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "simple_async_with_sync", "ctx": {"sample": "name"}},
        )

        self.assertTrue(res["is_queued"])

        res = mast.general_interface_to_api(
            api_name="walker_queue_wait",
            params={"task_id": res["result"], "timeout": 15},
        )

        self.assertEqual("test", res["result"]["anchor"])

        res = res["result"]["response"]

        # report from task1 (awaited)
        self.assertEqual(1, res["report"][0])
        self.assertEqual(2, res["report"][1])

        # task1 anchor
        self.assertEqual(1, res["report"][2])
        report4 = res["report"][4]

        res = res["report"][3]

        # no celery running
        self.assertTrue(res["is_queued"])

        res = mast.general_interface_to_api(
            api_name="walker_queue_check",
            params={"task_id": res["result"]},
        )

        self.assertEqual(report4, res)
        self.assertEqual(2, res["result"]["anchor"])
        self.assertEqual([2, 2], res["result"]["response"]["report"])

    def test_async_sync_syntax_without_celery(self):
        mast = JsOrc.master()
        JsOrc.svc("task").state = State.NOT_STARTED
        mast.sentinel_register(name="test", code=jtp.async_syntax, auto_run="")
        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "simple_async_with_sync", "ctx": {"sample": "name"}},
        )

        # no celery running
        self.assertFalse(res["is_queued"])

        # report from task1 (awaited)
        self.assertEqual(1, res["result"]["report"][0])
        self.assertEqual(2, res["result"]["report"][1])

        # report from task2 (async but no celery)
        self.assertEqual(2, res["result"]["report"][2])
        self.assertEqual(2, res["result"]["report"][3])

        # task1 anchor
        self.assertEqual(1, res["result"]["report"][4])

        res = res["result"]["report"][5]

        # no celery running
        self.assertFalse(res["is_queued"])
        self.assertEqual(2, res["result"])

        JsOrc.svc("task").state = State.RUNNING

    def test_block_scope_check(self):
        mast = JsOrc.master()
        mast.sentinel_register(name="test", code=jtp.block_scope_check, auto_run="")
        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(res["report"][-1], 10)

    def test_ignore_check(self):
        mast = JsOrc.master()
        res = mast.sentinel_register(name="test", code=jtp.ignore_check, auto_run="")
        res = mast.general_interface_to_api(
            api_name="walker_run", params={"name": "init"}
        )
        self.assertEqual(len(res["report"]), 9)

    @pytest.mark.order(1)
    def test_module_on_async(self):
        mast = JsOrc.super_master()
        if not JsOrc.svc("task").is_running():
            self.skip_test("Celery not running")

        with open("jaseci/tests/fixtures/non_existing_action.py", "w") as file:
            file.write(
                "from jaseci.jsorc.live_actions import jaseci_action\n@jaseci_action(act_group=['sim1'])\ndef tester():\n\treturn 1"
            )

        mast.sentinel_register(name="test", code=jtp.async_module, auto_run="")

        # sequence is relevant
        mast.actions_load_local("jaseci/tests/fixtures/non_existing_action.py")
        mast.actions_load_local("jaseci/tests/fixtures/existing_action.py")

        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "a", "ctx": {}},
        )

        self.assertTrue(res["success"])

        # will return [1,2] since all local actions are loaded on MainProcess
        self.assertEqual([1, 2], res["report"])

        os.remove("jaseci/tests/fixtures/non_existing_action.py")

        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "b", "ctx": {}},
        )

        self.assertTrue(res["is_queued"])

        res = mast.general_interface_to_api(
            api_name="walker_queue_wait",
            params={"task_id": res["result"], "timeout": 15},
        )

        self.assertEqual("SUCCESS", res["status"])

        # since sim1 file is already removed
        # upon running of task with sim1 it will try to rebuild
        # but file is already been deleted
        # without the fix it will throw error and will not continue on other action_sets
        # return would be [None, None] if not yet fixed
        self.assertEqual([None, 2], res["result"]["response"]["report"])

        self.assertIn(
            "Cannot execute sim1.tester - Not Found",
            res["result"]["response"]["errors"][0],
        )

    def test_walker_run_with_null_arguments(self):
        mast = JsOrc.master()
        res = mast.sentinel_register(
            name="test", code=jtp.walker_null_args, auto_run=""
        )
        res = mast.general_interface_to_api(
            api_name="walker_run",
            params={"name": "a", "ctx": {}},
        )
        self.assertTrue(res["success"])
        self.assertTrue(res["report"], [None, None])
        self.assertIsNone(res.get("errors"))



from jaseci.jsorc.live_actions import jaseci_action
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.mail_svc import Mailer


@jaseci_action()
def send(sender, recipients, subject, text, html):
    JsOrc.svc("mail").poke(Mailer).send_custom_email(
        sender, recipients, subject, (text, html)
    )



"""Built in actions for Jaseci"""
from jaseci.jsorc.live_actions import jaseci_action
from jaseci.utils.utils import print_stack_to_log
from base64 import b64decode, b64encode
import zlib


@jaseci_action()
def compress(data_b64: str = ""):
    """
    Compress data
    Param 1 - data in base64

    Return - compressed data in base64
    """
    data_bytes = b64decode(data_b64.encode("ascii"))
    data_compressed_bytes = zlib.compress(data_bytes)
    data_compressed_b64 = b64encode(data_compressed_bytes).decode("ascii")
    return data_compressed_b64


@jaseci_action()
def decompress(data_b64: str = ""):
    """
    Decompress data
    Param 1 - data in base64

    Return - decompressed data in base64
    """
    data_bytes = b64decode(data_b64.encode("ascii"))
    data_decompressed_bytes = zlib.decompress(data_bytes)
    data_decompressed_b64 = b64encode(data_decompressed_bytes).decode("ascii")
    return data_decompressed_b64



from jaseci.utils.test_core import CoreTest


class StdLibTest(CoreTest):
    fixture_src = __file__

    def test_clear_report(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("general.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "report_clearing"}])
        self.assertEqual(len(ret["report"]), 1)
        self.assertEqual(ret["report"][0], 7)

    def test_internal_lib(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("general.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "internal_lib"}])
        self.assertEqual(len(ret["report"]), 1)
        self.assertTrue(
            ret["report"][0].startswith("ncalls,tottime,percall,cumtime,percall,")
        )

    def test_rand_float_round(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("general.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "rand_float_round"}])
        self.assertEqual(ret["report"][0], 2.313)
        self.assertEqual(ret["report"][1], 2.847)



from jaseci.utils.test_core import CoreTest


class NetLibTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_pack(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("net_pack.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "pack_it"}])
        self.assertEqual(len(ret["report"][0]["nodes"]), 11)
        self.assertEqual(len(ret["report"][0]["edges"]), 10)

    def test_pack_subgraph(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("net_pack.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "pack_it_subgraph"}])
        self.assertEqual(len(ret["report"][0]["nodes"]), 5)
        self.assertEqual(len(ret["report"][0]["edges"]), 4)

    def test_pack_unpack(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("net_pack.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "pack_unpack"}])

        self.assertEqual(ret["report"][0], 16)

    def test_pack_unpack_terse(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("net_pack.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "pack_unpack_terse"}])
        self.assertEqual(ret["report"][0], 16)

    def test_pack_and_destroy(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("net_pack.jac")}],
        )
        before = len(self.smast._h.mem)
        ret = self.call(self.mast, ["walker_run", {"name": "pack_it_destroy"}])
        self.assertEqual(len(ret["report"]), 1)
        after = len(self.smast._h.mem)
        self.assertEqual(before, after)

    def test_pack_anc_priv(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("net_pack.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "pack_it_anc_priv"}])
        self.assertEqual(len(ret["report"]), 6)
        self.assertNotIn("priv", ret["report"][2]["context"])



from jaseci.utils.test_core import CoreTest, jac_testcase


class VectorTest(CoreTest):
    """UnitTest for Vector Module"""

    fixture_src = __file__

    @jac_testcase("vector.jac", "get_centroid_test")
    def test_get_centroid(self, ret):
        ret = [ret["report"][0][0], round(ret["report"][0][1], 3)]
        self.assertEqual(ret, [[1.0, 2.0, 3.5], 0.998])

    @jac_testcase("vector.jac", "softmax_test")
    def test_softmax(self, ret):
        ret = [round(num, 2) for num in ret["report"][0]]
        self.assertEqual(ret, [0.09, 0.24, 0.67])

    @jac_testcase("vector.jac", "dot_product_test")
    def test_dot_product(self, ret):
        self.assertEqual(ret["report"][0], 17)

    @jac_testcase("vector.jac", "cosine_similarity_test")
    def test_cosine_similarity_single(self, ret):
        ret = round(ret["report"][0], 3)
        self.assertEqual(ret, 0.991)

    @jac_testcase("vector.jac", "cosine_similarity_batch_test")
    def test_cosine_similarity_batch(self, ret):
        ret = [round(num, 2) for num in ret["report"][0]]
        self.assertEqual(ret, [0.99, 1.0])

    @jac_testcase("vector.jac", "dimensionality_reduction_test")
    def test_dimensionality_reduction(self, ret):
        ret = len(ret["report"][0][0])
        self.assertEqual(ret, 2)



from jaseci.jac.jsci_vm.op_codes import JsCmp, JsOp, JsType, type_map, cmp_op_map
from jaseci.jac.machine.machine_state import MachineState
from jaseci.jac.jsci_vm.inst_ptr import InstPtr, from_bytes
from jaseci.jac.machine.jac_value import JacValue
from jaseci.jac.jsci_vm.disasm import DisAsm


class Stack(object):
    def __init__(self):
        self._stk = []

    def stack_is_empty(self) -> bool:
        return len(self._stk) == 0

    def pop(self):
        if self.stack_is_empty():
            raise Exception("JaseciMachine stack is empty")
        return self._stk.pop()

    def push(self, value):
        self._stk.append(value)

    def peak_stack(self, count=0):
        if not count:
            print(list(reversed(self._stk)))
        else:
            print(list(reversed(self._stk))[0:count])


class VirtualMachine(MachineState, Stack, InstPtr):
    def __init__(self, **kwargs):
        Stack.__init__(self)
        InstPtr.__init__(self)
        MachineState.__init__(self, **kwargs)
        self._op = self.build_op_call()
        self._op_id = id(self)

    def reset_vm(self):
        Stack.__init__(self)
        InstPtr.__init__(self)
        if id(self) != self._op_id:
            self._op = self.build_op_call()

    def build_op_call(self):
        op_map = {}
        for op in JsOp:
            op_map[op] = getattr(self, f"op_{op.name}")
        return op_map

    def run_bytecode(self, bytecode):
        self.reset_vm()
        self._bytecode = bytearray(bytecode)
        try:
            while self._ip < len(self._bytecode):
                self._op[self._bytecode[self._ip]]()
                self._ip += 1
        except Exception as e:
            self.disassemble(print_out=False, log_out=True)
            raise e

    def disassemble(self, print_out=True, log_out=False):
        return DisAsm().disassemble(self._bytecode, print_out, log_out)

    def op_PUSH_SCOPE(self):  # noqa
        pass

    def op_POP_SCOPE(self):  # noqa
        pass

    def op_ADD(self):  # noqa
        val = self.pop()
        val.value = val.value + self.pop().value
        self.push(val)

    def op_SUBTRACT(self):  # noqa
        val = self.pop()
        val.value = val.value - self.pop().value
        self.push(val)

    def op_MULTIPLY(self):  # noqa
        val = self.pop()
        val.value = val.value * self.pop().value
        self.push(val)

    def op_DIVIDE(self):  # noqa
        val = self.pop()
        val.value = val.value / self.pop().value
        self.push(val)

    def op_MODULO(self):  # noqa
        val = self.pop()
        val.value = val.value % self.pop().value
        self.push(val)

    def op_POWER(self):  # noqa
        val = self.pop()
        val.value = val.value ** self.pop().value
        self.push(val)

    def op_NEGATE(self):  # noqa
        val = self.pop()
        val.value = -(val.value)
        self.push(val)

    def op_COMPARE(self):  # noqa
        ctyp = JsCmp(self.offset(1))
        val = self.pop()
        val.value = (
            cmp_op_map[ctyp](val.value, self.pop().value)
            if ctyp != JsCmp.NOT
            else cmp_op_map[ctyp](val.value)
        )
        self.push(val)
        self._ip += 1

    def op_AND(self):  # noqa
        val = self.pop()
        val.value = val.value and self.pop().value
        self.push(val)

    def op_OR(self):  # noqa
        val = self.pop()
        val.value = val.value or self.pop().value
        self.push(val)

    def op_ASSIGN(self):  # noqa
        dest = self.pop()
        self.perform_assignment(dest, self.pop())

    def op_COPY_FIELDS(self):  # noqa
        dest = self.pop()
        self.perform_copy_fields(dest, self.pop())

    def op_INCREMENT(self):  # noqa
        ityp = JsCmp(self.offset(1))
        dest = self.pop()
        self.perform_increment(dest, self.pop(), ityp)
        self._ip += 1

    def op_LOAD_CONST(self):  # noqa
        typ = JsType(self.offset(1))
        operand2 = self.offset(2)
        if typ in [JsType.TYPE]:
            val = type_map[JsType(operand2)]
            self._ip += 2
        elif typ in [JsType.INT]:
            val = from_bytes(type_map[typ], self.offset(3, operand2))
            self._ip += 2 + operand2
        elif typ in [JsType.STRING]:
            str_len = from_bytes(type_map[JsType.INT], self.offset(3, operand2))
            val = from_bytes(type_map[typ], self.offset(3 + operand2, str_len))
            self._ip += 2 + operand2 + str_len
        elif typ in [JsType.FLOAT]:
            val = from_bytes(float, self.offset(2, 8))
            self._ip += 1 + 8
        elif typ in [JsType.BOOL]:
            val = bool(self.offset(2))
            self._ip += 1 + 1
        self.push(JacValue(self, value=val))

    def op_LOAD_VAR(self):  # noqa
        name = from_bytes(str, self.offset(2, self.offset(1)))
        self.load_variable(name)
        self._ip += 1 + self.offset(1)

    def op_CREATE_VAR(self):  # noqa
        name = from_bytes(str, self.offset(2, self.offset(1)))
        self.load_variable(name, assign_mode=True)
        self._ip += 1 + self.offset(1)

    def op_REPORT(self):  # noqa
        self.report.append(self.pop())

    def op_ACTION_CALL(self):  # noqa
        pass

    def op_DEBUG_INFO(self):  # noqa
        byte_len_l = self.offset(1)
        line = from_bytes(int, self.offset(2, byte_len_l))
        f_offset = byte_len_l + 2
        byte_len_f = self.offset(f_offset)
        jacfile = (
            from_bytes(str, self.offset(f_offset + 1, byte_len_f))
            if byte_len_f
            else self._cur_jac_ast.loc[2]
        )
        self._ip += 2 + byte_len_l + byte_len_f
        self._cur_jac_ast.loc = [line, 0, jacfile, {}]



from os.path import dirname, join


def get_ver():
    with open(join(dirname(__file__), "VERSION")) as version_file:
        return version_file.read().strip()


__version__ = get_ver()
__creator__ = "Jason Mars and contributors"
__url__ = "https://jaseci.org"


def load_standard():
    import jaseci.extens.act_lib.net  # noqa
    import jaseci.extens.act_lib.rand  # noqa
    import jaseci.extens.act_lib.request  # noqa
    import jaseci.extens.act_lib.std  # noqa
    import jaseci.extens.act_lib.file  # noqa
    import jaseci.extens.act_lib.vector  # noqa
    import jaseci.extens.act_lib.date  # noqa
    import jaseci.extens.act_lib.jaseci  # noqa
    import jaseci.extens.act_lib.mail  # noqa
    import jaseci.extens.act_lib.task  # noqa
    import jaseci.extens.act_lib.internal  # noqa
    import jaseci.extens.act_lib.zip  # noqa
    import jaseci.extens.act_lib.webtool  # noqa
    import jaseci.extens.act_lib.elastic  # noqa
    import jaseci.extens.act_lib.url  # noqa
    import jaseci.extens.act_lib.stripe  # noqa
    import jaseci.extens.act_lib.regex  # noqa


load_standard()



from jaseci.utils.test_core import CoreTest


class TestCodegen(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_simple_codegen(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("simple.jac")}],
        )
        # self.log(ret)
        ret = self.call(self.mast, ["walker_run", {"name": "most_basic"}])
        # self.log(ret)
        # from jaseci.jac.ir.passes import PrinterPass
        # PrinterPass(self.mast.active_snt()._jac_ast).run()
        self.assertEqual(ret["report"][0], 5004)
        self.assertEqual(ret["report"][1], 1)
        self.assertEqual(ret["report"][2], "ab")
        self.assertEqual(ret["report"][3], {"base": 5, "test": 5})
        self.assertEqual(ret["report"][4], {"a": 3, "b": 55, "c": 311})

    def test_unicode(self):
        ret = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("unicode.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "unicode"}])
        self.assertEqual(len(ret["report"]), 1)



from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.elastic_svc import ElasticService
from jaseci.utils.test_core import CoreTest


class ElasticTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def setUp(self):
        super().setUp()

    @JsOrc.inject(services=["elastic"])
    def test_elastic(self, elastic: ElasticService):
        if not elastic.is_running():
            self.skipTest("Elastic is not available!")

        res = self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("elastic.jac")}],
        )

        res = self.call(
            self.mast,
            [
                "walker_run",
                {"name": "test_activity", "ctx": {"a": 1, "b": 2, "c": 3}},
            ],
        )

        # report std.log_activity({"other_field": 1}, action = "CUSTOM_ACTION_NAME");
        self.assertEqual("created", res["report"][0]["result"])

        # report std.log_activity({"other_field": 2}, action = "CUSTOM_ACTION_NAME", suffix="-sample-id");
        self.assertEqual("created", res["report"][1]["result"])

        # report elastic.doc_activity({"testing": "doc_activity"});
        self.assertEqual("created", res["report"][2]["result"])

        # report elastic.search_activity({});
        hits = res["report"][3]["hits"]

        self.assertEqual(2, hits["total"]["value"])
        self.assertEqual({"a": 1, "b": 2, "c": 3}, hits["hits"][0]["_source"]["data"])
        self.assertEqual(1, hits["hits"][0]["_source"].get("other_field"))
        self.assertEqual({"testing": "doc_activity"}, hits["hits"][1]["_source"])

        # report elastic.search_activity({}, suffix="-sample-id");
        hits = res["report"][4]["hits"]

        self.assertEqual(1, hits["total"]["value"])
        self.assertEqual({"a": 1, "b": 2, "c": 3}, hits["hits"][0]["_source"]["data"])
        self.assertEqual(2, hits["hits"][0]["_source"].get("other_field"))

        # report elastic.mapping_activity();
        mapping: dict = res["report"][5]
        for key, value in mapping.items():
            self.assertTrue(key.endswith("-activity"))
            self.assertTrue(key.startswith("jaseci-"))
            self.assertTrue(value["mappings"])

        # report elastic.mapping_activity(suffix="-sample-id");
        mapping: dict = res["report"][6]
        for key, value in mapping.items():
            self.assertTrue(key.endswith("-activity-sample-id"))
            self.assertTrue(key.startswith("jaseci-"))
            self.assertTrue(value["mappings"])

        res = self.call(
            self.mast,
            [
                "walker_run",
                {"name": "test_common", "ctx": {"d": 1, "e": 2, "f": 3}},
            ],
        )

        # report elastic.doc({"testing": "doc"});
        self.assertEqual("created", res["report"][0]["result"])

        # elastic.doc({"testing": "doc-other-index"}, index="other-index");
        self.assertEqual("created", res["report"][1]["result"])

        # elastic.doc({"testing": "doc-other-index"}, index="other-index", suffix="-user-id");
        self.assertEqual("created", res["report"][2]["result"])

        # report elastic.search({});
        hits = res["report"][3]["hits"]

        self.assertEqual(1, hits["total"]["value"])
        self.assertEqual({"testing": "doc"}, hits["hits"][0]["_source"])

        # elastic.search({}, index="other-index");
        hits = res["report"][4]["hits"]

        self.assertEqual(1, hits["total"]["value"])
        self.assertEqual({"testing": "doc-other-index"}, hits["hits"][0]["_source"])

        # elastic.search({}, index="other-index", suffix="-user-id");
        hits = res["report"][5]["hits"]

        self.assertEqual(1, hits["total"]["value"])
        self.assertEqual(
            {"testing": "doc-other-index-suffix"}, hits["hits"][0]["_source"]
        )

        # report elastic.mapping();
        mapping: dict = res["report"][6]
        for key, value in mapping.items():
            self.assertTrue(key.endswith("-common"))
            self.assertTrue(key.startswith("jaseci-"))
            self.assertTrue(value["mappings"])

        # report elastic.mapping(index="other-index");
        mapping: dict = res["report"][7]
        for key, value in mapping.items():
            self.assertEqual(key, "other-index")
            self.assertTrue(value["mappings"])

        # report elastic.mapping(index="other-index", suffix="-user-id");
        mapping: dict = res["report"][8]
        for key, value in mapping.items():
            self.assertEqual(key, "other-index-user-id")
            self.assertTrue(value["mappings"])

        res = self.call(
            self.mast,
            [
                "walker_run",
                {"name": "test_other", "ctx": {}},
            ],
        )

        # report elastic.reindex
        self.assertEqual(1, res["report"][0]["created"])

        # report elastic.aliases();
        aliases: dict = res["report"][1]
        self.assertEqual(6, len(aliases.keys()))
        self.assertTrue(aliases.pop("other-index"))
        self.assertTrue(aliases.pop("other-index-user-id"))
        self.assertTrue(aliases.pop("new-index"))
        for key in list(aliases.keys()):
            if key.startswith("jaseci-"):
                if (
                    key.endswith("-activity")
                    or key.endswith("-common")
                    or key.endswith("-activity-sample-id")
                ):
                    self.assertTrue(aliases.pop(key))
        self.assertFalse(aliases)

        # report elastic.search({}, index="new-index");
        hits = res["report"][2]["hits"]

        self.assertEqual(1, hits["total"]["value"])
        self.assertEqual(
            {"testing": "doc-other-index-suffix"}, hits["hits"][0]["_source"]
        )



from jaseci.prim.edge import Edge
from jaseci.prim.node import Node
from jaseci.utils.test_core import CoreTest


class WalkerApiTest(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_walker_yield(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], ["entering", 7])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [8])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [9])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [10])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], ["should start over now", "exiting"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], ["entering", 7])

    def test_walker_yield_update(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], ["entering", 7])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [8])
        ret = self.call(
            self.mast, ["walker_run", {"name": "test_yield", "ctx": {"a": 2}}]
        )
        self.assertEqual(ret["report"], [3])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [4])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [5])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [6])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [7])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [8])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [9])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], [10])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], ["should start over now", "exiting"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield"}])
        self.assertEqual(ret["report"], ["entering", 7])

    def test_walker_yield_report(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_report"}])
        self.assertEqual(ret["report"], [7])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_report"}])
        self.assertEqual(ret["report"], [8])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_report"}])
        self.assertEqual(ret["report"], [9])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_report"}])
        self.assertEqual(ret["report"], [10])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_report"}])
        self.assertEqual(ret["report"], ["should start over now"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_report"}])
        self.assertEqual(ret["report"], [7])

    def test_walker_yield_disengage(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_disengage"}])
        self.assertEqual(ret["report"], [7])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_disengage"}])
        self.assertEqual(ret["report"], [8])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_disengage"}])
        self.assertEqual(ret["report"], [9])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_disengage"}])
        self.assertEqual(ret["report"], [10])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_disengage"}])
        self.assertEqual(ret["report"], ["should start over now"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_disengage"}])
        self.assertEqual(ret["report"], [7])

    def test_walker_yield_take(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_take"}])
        self.assertEqual(ret["report"], [7, "test"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_take"}])
        self.assertEqual(ret["report"], [8, "test"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_take"}])
        self.assertEqual(ret["report"], [9, "test"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_take"}])
        self.assertEqual(ret["report"], [10, "test"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_take"}])
        self.assertEqual(ret["report"], ["should start over now"])
        ret = self.call(self.mast, ["walker_run", {"name": "test_yield_take"}])
        self.assertEqual(ret["report"], [7, "test"])

    def test_walker_smart_yield(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "smart_yield"}])
        before = self.mast._h.get_object_distribution()
        ret = self.call(self.mast, ["walker_run", {"name": "smart_yield"}])
        ret = self.call(self.mast, ["walker_run", {"name": "smart_yield"}])
        ret = self.call(self.mast, ["walker_run", {"name": "smart_yield"}])
        after = self.mast._h.get_object_distribution()
        self.assertEqual(before, after)
        self.assertEqual(ret["report"], [{"id": 2}])

    def test_walker_smart_yield_no_future(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "smart_yield_no_future"}])
        before = self.mast._h.get_object_distribution()
        ret = self.call(self.mast, ["walker_run", {"name": "smart_yield_no_future"}])
        ret = self.call(self.mast, ["walker_run", {"name": "smart_yield_no_future"}])
        ret = self.call(self.mast, ["walker_run", {"name": "smart_yield_no_future"}])
        after = self.mast._h.get_object_distribution()
        self.assertEqual(before, after)
        self.assertEqual(ret["report"], [{}])

    def test_walker_deep_yield(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "deep_yield"}])
        self.assertEqual(
            ret["report"],
            [
                {},
                {"id": 0},
                {"id": 1},
                {"id": 2},
                {"id": 3},
                {"id": 4},
                {"id": 5},
                {"id": 6},
                {"id": 7},
                {"id": 8},
                {"id": 9},
                {},
                {"id": 0},
                {"id": 1},
                {"id": 2},
                11,
            ],
        )

    def test_walker_simple_yield(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        expected = [
            "entry",
            {"id": 0},
            {"id": 1},
            {"id": 2},
            {"id": 3},
            {"id": 4},
            {"id": 5},
            {"id": 6},
            {"id": 7},
            {"id": 8},
            {"id": 9},
            "entry",
            {"id": 0},
            {"id": 10},
            {"id": 1},
            {"id": 11},
        ]
        for i in range(16):
            ret = self.call(self.mast, ["walker_run", {"name": "simple_yield"}])
            self.assertEqual(ret["report"][0], expected[i])

    def test_walker_deep_yield2(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "deep_yield2"}])
        self.assertEqual(
            ret["report"],
            [
                "entry",
                {},
                {"id": 0},
                {"id": 1},
                {"id": 2},
                {"id": 3},
                {"id": 4},
                {"id": 5},
                {"id": 6},
                {"id": 7},
                {"id": 8},
                {"id": 9},
                "entry",
                {},
                {"id": 0},
                {"id": 10},
                {"id": 1},
                {"id": 11},
            ],
        )

    def test_walker_deep_yield_no_leak(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        before = self.mast._h.get_object_distribution()
        self.call(self.mast, ["walker_run", {"name": "deep_yield"}])
        self.call(self.mast, ["walker_run", {"name": "deep_yield"}])
        self.call(self.mast, ["walker_run", {"name": "deep_yield"}])
        after = self.mast._h.get_object_distribution()
        after.pop(Node)
        after.pop(Edge)
        self.assertEqual(before, after)

    def test_walker_simple_yield_skip_test(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = []
        for i in range(16):
            ret += self.call(
                self.mast, ["walker_run", {"name": "simple_yield_skip_test"}]
            )["report"]
        self.assertEqual(
            ret,
            [
                "entry",
                {},
                "in_node",
                {"id": 0},
                "in_node",
                {"id": 1},
                "in_node",
                {"id": 2},
                "in_node",
                {"id": 3},
                "in_node",
                {"id": 4},
                5,
                {"id": 5},
                5,
                {"id": 6},
                5,
                {"id": 7},
                5,
                {"id": 8},
                5,
                {"id": 9},
                "entry",
                {},
                "in_node",
                {"id": 0},
                "in_node",
                {"id": 10},
                "in_node",
                {"id": 1},
                "in_node",
                {"id": 11},
            ],
        )

    def test_error_reporting_walker_only_actions(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("walker_yield.jac")}],
        )
        ret = []
        self.call(self.mast, ["walker_run", {"name": "error_walker_action"}])
        ret = self.call(self.mast, ["walker_run", {"name": "error_walker_action"}])
        self.assertIn('cannot execute the statement "disengage ; "', ret["errors"][0])

    def test_walker_stepping(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("fam.jac")}],
        )
        ret = self.call(self.mast, ["walker_spawn_create", {"name": "create_fam"}])
        jid = ret["jid"]
        ret = self.call(self.mast, ["walker_prime", {"wlk": jid}])
        ret = self.call(self.mast, ["walker_step", {"wlk": jid}])
        self.log(ret)
        self.assertEqual(len(ret["next_node_ids"]), 2)
        ret = self.call(self.mast, ["walker_step", {"wlk": jid}])
        self.assertEqual(len(ret["next_node_ids"]), 1)
        ret = self.call(self.mast, ["walker_step", {"wlk": jid}])
        self.assertEqual(len(ret["next_node_ids"]), 0)
        ret = self.call(self.mast, ["walker_step", {"wlk": jid}])
        self.assertEqual(len(ret["next_node_ids"]), 0)



from jaseci.utils.test_core import CoreTest, jac_testcase


class WebtoolTest(CoreTest):
    """UnitTest for Webtool"""

    fixture_src = __file__

    @jac_testcase("webtool.jac", "get_meta_valid")
    def test_get_meta_valid(self, ret):
        self.assertTrue(ret["success"])
        expected_tags = set(["og:image", "og:type", "og:title"])
        tags = set(
            [
                meta["property"] if "property" in meta else ""
                for meta in ret["report"][0]
            ]
        )
        self.assertTrue(tags.issuperset(expected_tags))

    @jac_testcase("webtool.jac", "get_meta_need_auth")
    def test_get_meta_need_auth(self, ret):
        self.assertTrue(ret["success"])
        self.assertTrue(len(ret["report"][0]) > 0)

    @jac_testcase("webtool.jac", "get_meta_invalid")
    def test_get_meta_invalid(self, ret):
        self.assertTrue("Failed at getting metadata" in ret["report"][0])



import stripe
from unittest.mock import Mock
from jaseci.utils.test_core import CoreTest, jac_testcase
from jaseci.jsorc.jsorc import JsOrc


class StripeTests(CoreTest):
    """Unit tests for Stripe actions"""

    fixture_src = __file__

    @classmethod
    def setUpClass(cls):
        config = JsOrc.settings("STRIPE_CONFIG")
        config["enabled"] = True
        config[
            "api_key"
        ] = "sk_test_51JWUIeCZO78n7fsZnPvualWhmJg1DcCI332kKnWF3q2sKGwnPADjEmNblfFWi4pWAWPuJwHxpeSoJGc0J5ButHN900Q2xBz1se"
        config["webhook_key"] = "test_webhook_key"

        JsOrc.svc_reset("stripe")

        super(StripeTests, cls).setUpClass()
        stripe.Product.create = Mock()
        stripe.Price.create = Mock()
        stripe.Product.list = Mock()
        stripe.Customer.create = Mock()
        stripe.Customer.retrieve = Mock()
        stripe.PaymentMethod.list = Mock()
        stripe.PaymentMethod.attach = Mock()
        stripe.PaymentMethod.detach = Mock()
        stripe.Customer.modify = Mock()
        stripe.Invoice.create = Mock()
        stripe.Invoice.list = Mock()
        stripe.PaymentIntent.list = Mock()
        stripe.PaymentIntent.create = Mock()
        stripe.Subscription.list = Mock()
        stripe.PaymentMethod.create = Mock()
        stripe.Subscription.create = Mock()
        stripe.Subscription.delete = Mock()
        stripe.Subscription.retrieve = Mock()
        stripe.Subscription.modify = Mock()
        stripe.Invoice.retrieve = Mock()
        stripe.SubscriptionItem.create_usage_record = Mock()
        stripe.checkout.Session.create = Mock()

    @classmethod
    def tearDownClass(cls):
        super(StripeTests, cls).tearDownClass()

    @jac_testcase("stripe.jac", "create_product")
    def test_stripe_create_product(self, ret):
        stripe.Product.create.assert_called_once_with(
            name="product1", description="new product"
        )

    @jac_testcase("stripe.jac", "create_product_price")
    def test_stripe_create_product_price(self, ret):
        stripe.Price.create.assert_called_once_with(
            product="product1", unit_amount=12, currency="usd"
        )

    @jac_testcase("stripe.jac", "product_list")
    def test_stripe_product_list(self, ret):
        stripe.Product.list.assert_called_once_with(active=True)

    @jac_testcase("stripe.jac", "create_customer")
    def test_stripe_create_customer(self, ret):
        stripe.Customer.create.assert_called_once_with(
            email="test12@gmail.com",
            name="stripe customer",
            address={"billing_address": "123 metro manila"},
        )

    @jac_testcase("stripe.jac", "get_customer")
    def test_stripe_get_customer(self, ret):
        stripe.Customer.retrieve.assert_called_once_with(id="cus_NBsqL1C1GrrHYM")

    @jac_testcase("stripe.jac", "attach_payment_method")
    def test_stripe_attach_payment_method(self, ret):
        stripe.PaymentMethod.list.assert_called_once_with(
            customer="cus_NBsqL1C1GrrHYM", type="card"
        )

        stripe.PaymentMethod.attach.assert_called_once_with(
            payment_method="pm_1MN1iN2xToAoV8chTjvX94hm", customer="cus_NBsqL1C1GrrHYM"
        )

    @jac_testcase("stripe.jac", "detach_payment_method")
    def test_stripe_detach_payment_method(self, ret):
        stripe.PaymentMethod.detach.assert_called_once_with(
            payment_method="pm_1MN1iN2xToAoV8chTjvX94hm"
        )

    @jac_testcase("stripe.jac", "get_payment_methods")
    def test_stripe_get_payment_methods(self, ret):
        stripe.PaymentMethod.list.assert_called_with(
            customer="cus_NBsqL1C1GrrHYM", type="card"
        )

    @jac_testcase("stripe.jac", "update_default_payment_method")
    def test_stripe_update_default_payment_method(self, ret):
        stripe.Customer.modify.assert_called_with(
            sid="cus_NBsqL1C1GrrHYM",
            invoice_settings={"default_payment_method": "pm_1MN1iN2xToAoV8chTjvX94hm"},
        )

    @jac_testcase("stripe.jac", "create_invoice")
    def test_stripe_create_invoice(self, ret):
        stripe.Customer.modify.assert_called_once_with(
            sid="cus_NBsqL1C1GrrHYM",
            invoice_settings={"default_payment_method": "pm_1MN1iN2xToAoV8chTjvX94hm"},
        )

    @jac_testcase("stripe.jac", "get_invoice_list")
    def test_stripe_get_invoice_list(self, ret):
        stripe.Invoice.list.assert_called_once_with(
            customer="cus_NBsqL1C1GrrHYM",
            limit=10,
            subscription="sub_1MTgMQCZO78n7fsZqu1dk6nD",
        )

    @jac_testcase("stripe.jac", "get_payment_intents")
    def test_stripe_get_payment_intents(self, ret):
        stripe.PaymentIntent.list.assert_called_once_with(
            customer="cus_NBsqL1C1GrrHYM", limit=10
        )

    @jac_testcase("stripe.jac", "create_payment_intents")
    def test_stripe_create_payment_intents(self, ret):
        stripe.PaymentIntent.create.assert_called_once_with(
            customer="cus_NBsqL1C1GrrHYM",
            amount=12,
            currency="usd",
            payment_method_types=["card"],
        )

    @jac_testcase("stripe.jac", "get_customer_subscription")
    def test_stripe_get_customer_subscription(self, ret):
        stripe.Subscription.list.assert_called_once_with(
            customer="sub_1MTgMQCZO78n7fsZqu1dk6nD"
        )

    @jac_testcase("stripe.jac", "create_payment_method")
    def test_stripe_create_payment_method(self, ret):
        stripe.PaymentMethod.create.assert_called_once_with(
            type="card",
            card={
                "number": "4242424242424242",
                "exp_month": 8,
                "exp_year": 2024,
                "cvc": "314",
            },
            billing_details={"city": "Caloocan", "country": "philippines"},
        )

    @jac_testcase("stripe.jac", "create_trial_subscription")
    def test_stripe_create_trial_subscription(self, ret):
        stripe.Subscription.create.assert_called_with(
            customer="cus_NBsqL1C1GrrHYM",
            items=[{"price": "price_1MR9T6CZO78n7fsZmNdIJplr"}],
            trial_period_days=14,
        )

    @jac_testcase("stripe.jac", "create_subscription")
    def test_stripe_create_subscription(self, ret):
        stripe.Subscription.create.assert_called_with(
            customer="cus_NBsqL1C1GrrHYM",
            items=[{"price": "price_1MR9T6CZO78n7fsZmNdIJplr"}],
        )

    @jac_testcase("stripe.jac", "cancel_subscription")
    def test_stripe_cancel_subscription(self, ret):
        stripe.Subscription.delete.assert_called_once_with(
            sid="sub_1MTgMQCZO78n7fsZqu1dk6nD"
        )

    @jac_testcase("stripe.jac", "get_subscription")
    def test_stripe_get_subscription(self, ret):
        stripe.Subscription.retrieve.assert_called_once_with(
            id="sub_1MTgMQCZO78n7fsZqu1dk6nD"
        )

    @jac_testcase("stripe.jac", "update_subscription_item")
    def test_stripe_update_subscription_item(self, ret):
        stripe.Subscription.modify.assert_called_once_with(
            sid="sub_1MTgMQCZO78n7fsZqu1dk6nD",
            cancel_at_period_end=False,
            items=[
                {
                    "id": "su_1MTgMQCZO78n7fsZqu1dk6nD",
                    "price": "price_1MTgMQCZO78n7fsZqu1dk6nD",
                }
            ],
        )

    @jac_testcase("stripe.jac", "get_invoice")
    def test_stripe_get_invoice(self, ret):
        stripe.Invoice.retrieve.assert_called_once_with(
            id="inv_1MTgMQCZO78n7fsZqu1dk6nD"
        )

    @jac_testcase("stripe.jac", "create_usage_report")
    def test_stripe_create_usage_report(self, ret):
        stripe.SubscriptionItem.create_usage_record.assert_called()

    @jac_testcase("stripe.jac", "create_checkout_session")
    def test_stripe_create_checkout_session(self, ret):
        stripe.checkout.Session.create.assert_called_once_with(
            success_url="https://example.com/success",
            cancel_url="https://example.com/cancel",
            line_items=[{"price": "price_H5ggYwtDq4fbrJ", "quantity": 12}],
            mode="payment",
        )



"""
Variable scope manager for Jac

Utility for all runtime interaction with variables in different scopes
"""
from jaseci.utils.id_list import IdList
from jaseci.jac.machine.jac_value import JacValue
from jaseci.jsorc.live_actions import get_global_actions

global_action_sets = None


class JacScope:
    def __init__(self, parent, has_obj, action_sets):
        self.parent = parent
        self.local_scope = {}
        self.has_obj = has_obj if has_obj else self
        self.context = {}
        self.action_sets = action_sets
        self.setup_actions()

    def setup_actions(self):
        global global_action_sets
        if global_action_sets is None:
            global_action_sets = self.group_actions(get_global_actions())
        allactions = []
        for i in self.action_sets:
            allactions += i.obj_list()
        self.action_sets = global_action_sets
        for i in allactions:
            self.add_action(i)

    def add_action(self, act):
        group = act.name.split(".")[0]
        if "." in act.name:
            if group not in self.action_sets.keys():
                self.action_sets[group] = {}
            action = act.name.split(".")[1]
            self.action_sets[group][action] = act
        else:
            self.action_sets[group] = act

    def group_actions(self, act_list):
        action_sets = {}
        for act in act_list:
            group = act.name.split(".")[0]
            if "." in act.name:
                if group not in action_sets.keys():
                    action_sets[group] = {}
                action = act.name.split(".")[1]
                action_sets[group][action] = act
            else:
                action_sets[group] = act
        return action_sets

    def set_agent_refs(self, cur_node, cur_walker):
        self.local_scope["here"] = cur_node
        self.local_scope["visitor"] = cur_walker

    def inherit_agent_refs(self, src_scope, src_node):  # used for calls of abilities
        self.local_scope["here"] = src_node
        self.local_scope["visitor"] = src_scope.local_scope["visitor"]

    def get_aganet_refs(self):
        return {
            "here": self.local_scope["here"],
            "visitor": self.local_scope["visitor"],
        }

    def find_live_attr(self, name, allow_read_only=True):
        """Finds binding for variable if not in standard scope"""
        # check if var is in walker's context
        if "context" in self.has_obj.__dict__ and name in self.has_obj.context.keys():
            return JacValue(self.parent, ctx=self.has_obj, name=name)
        elif name in self.action_sets.keys():
            return JacValue(self.parent, ctx=self.action_sets, name=name)
        return None

    def get_live_var(self, name, create_mode=False):
        """Returns live variable"""
        found = None
        # Lock for variable in various locations
        if name in self.local_scope.keys():
            found = JacValue(self.parent, ctx=self.local_scope, name=name)
        else:
            found = self.find_live_attr(name)
        if found is None and create_mode:
            self.local_scope[name] = None
            return JacValue(self.parent, ctx=self.local_scope, name=name)
        if found:
            return found
        return None



import os
import unittest
from unittest import TestCase

from jaseci.jsorc.jsorc import JsOrc
from jaseci.utils.utils import TestCaseHelper


class CoreTest(TestCaseHelper, TestCase):
    """Unit tests for Jac Core APIs"""

    fixture_src = __file__

    def setUp(self):
        super().setUp()
        self.smast = JsOrc.super_master()
        self.mast = JsOrc.master(h=self.smast._h)

    def tearDown(self):
        super().tearDown()

    def call(self, mast, pl):
        ret = mast.general_interface_to_api(api_name=pl[0], params=pl[1])
        return ret

    def load_jac(self, fn):
        with open(os.path.dirname(self.fixture_src) + "/fixtures/" + fn) as f:
            return f.read()


def jac_testcase(jac_file: str, test_name: str, log_out: bool = False):
    """decorator for test cases"""

    def decorator(func):
        def wrapper(self):
            if log_out:
                self.logger_on()
            self.call(
                self.mast,
                ["sentinel_register", {"code": self.load_jac(jac_file)}],
            )
            ret = self.call(self.mast, ["walker_run", {"name": test_name}])
            func(self, ret)
            if log_out:
                self.logger_off()

        return wrapper

    return decorator


def skip_without_redis(test):
    """
    Skip test if expected not to work without redis.
    """

    def skipper(*args, **kwargs):
        if not JsOrc.svc("redis").is_running():
            raise unittest.SkipTest("No Redis!")
        test(*args, **kwargs)

    return skipper



from jaseci.jac.ir.passes.ir_pass import IrPass
import sys


class StatsPass(IrPass):
    def __init__(self, *args):
        super().__init__(*args)
        self.stats = {"node_count": 0, "ir_size_kb": 0}

    def enter_node(self, node):
        self.stats["node_count"] += 1
        self.stats["ir_size_kb"] += sys.getsizeof(node) / 1024

    def after_pass(self):
        self.stats["ir_size_kb"] = round(self.stats["ir_size_kb"], 2)
        print(self.stats, "\n")



class ActionsState:
    """
    An object tracking the various states of available actions. Utilized by the ActionsOptimizer and ActionsOptimizerPolicy
    """

    def __init__(self):
        """
        Actions state format:
        actions_state = {
            "ACTION_NAME": {
                "mode": "local|remote",
                "module": {
                    "name": MDOULE_NAME
                },
                "local": {
                    "path": LOCAL MODULE FILE PATH
                },
                "remote": {
                    "url": "remote_url of the action microservice",
                    "status": "READY|STARTING|RETIRING"
                }
            }
        }
        """
        self.state = {}

    def init_state(self, name):
        """
        Initialize the data structure for the action state tracking for a specific action
        """
        self.state[name] = {
            "mode": None,
            "module": {"name": None, "loaded_module": None},
            "local": {"path": None},
            "remote": {"url": None, "status": None},
        }

        return self.state[name]

    def get_active_actions(self):
        """
        Return list of action modules currently used
        """
        return list(self.state.keys())

    def local_action_loaded(self, name):
        self.state[name]["mode"] = "local"

    def module_action_loaded(self, name, module, loaded):
        self.state[name]["mode"] = "module"
        self.state[name]["module"]["name"] = module
        self.state[name]["module"]["loaded_module"] = loaded

    def module_action_unloaded(self, name):
        self.state[name]["mode"] = None
        self.state[name]["module"]["name"] = None

    def remote_action_unloaded(self, name):
        self.state[name]["mode"] = None

    def remove_remote(self, name):
        self.state[name]["remote"] = {"url": None, "status": None}

    def remote_action_loaded(self, name):
        self.state[name]["mode"] = "remote"

    def start_remote_service(self, name, url):
        self.state[name]["remote"] = {"status": "STARTING", "url": url}

    def set_remote_action_ready(self, name):
        self.state[name]["remote"]["status"] = "READY"

    def stop_remote_action(self, name):
        self.state[name]["remote"] = {}

    def get_remote_url(self, name):
        if name not in self.state or "remote" not in self.state[name]:
            return None
        return self.state[name]["remote"].get("url", None)

    def get_all_state(self):
        ret = {}
        for name, state in self.state.items():
            ret[name] = {"mode": state["mode"]}
        return ret

    def get_state(self, name):
        return self.state.get(name, None)

    def set_url(self, name, url):
        if name not in self.state:
            self.state[name] = {}
        if "remote" not in self.state[name]:
            self.state[name]["remote"] = {}
        self.state[name]["remote"]["url"] = url



import ssl
from json import dumps
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from smtplib import SMTP, SMTP_SSL

from jaseci.jsorc.jsorc import JsOrc


#################################################
#                   EMAIL APP                   #
#################################################


@JsOrc.service(name="mail", config="MAIL_CONFIG")
class MailService(JsOrc.CommonService):
    ###################################################
    #                     BUILDER                     #
    ###################################################

    def run(self):
        self.__convert_config()
        self.app = self.connect()

    # ----------- BACKWARD COMPATIBILITY ------------ #
    # ---------------- TO BE REMOVED ---------------- #

    def __convert(self, hook, holder, mapping: dict):
        for k, v in mapping.items():
            if hook.has_glob(k):
                conf = hook.get_glob(k)
                if not (conf is None):
                    if v == "tls":
                        holder[v] = conf.lower() == "true"
                    else:
                        holder[v] = conf
        return holder

    def __convert_config(self):
        hook = JsOrc.hook()
        version = self.config.get("version", 2)
        migrate = self.config.get("migrate", False)
        if version == 1 or migrate:
            self.__convert(
                hook,
                self.config,
                {
                    "EMAIL_BACKEND": "backend",
                    "EMAIL_HOST": "host",
                    "EMAIL_PORT": "port",
                    "EMAIL_HOST_USER": "user",
                    "EMAIL_HOST_PASSWORD": "pass",
                    "EMAIL_DEFAULT_FROM": "sender",
                    "EMAIL_USE_TLS": "tls",
                },
            )

            if "templates" not in self.config:
                self.config["templates"] = {}

            self.__convert(
                hook,
                self.config["templates"],
                {
                    "EMAIL_ACTIVATION_SUBJ": "activation_subj",
                    "EMAIL_ACTIVATION_BODY": "activation_body",
                    "EMAIL_ACTIVATION_HTML_BODY": "activation_html_body",
                    "EMAIL_RESETPASS_SUBJ": "resetpass_subj",
                    "EMAIL_RESETPASS_BODY": "resetpass_body",
                    "EMAIL_RESETPASS_HTML_BODY": "resetpass_html_body",
                },
            )

            if migrate:
                self.config["migrate"] = False
                hook.save_glob("MAIL_CONFIG", dumps(self.config))

    ###################################################
    #                     CLEANER                     #
    ###################################################

    # ---------------- PROXY EVENTS ----------------- #

    def on_delete(self):
        if self.is_running():
            self.app.terminate()

    ####################################################
    #                    OVERRIDDEN                    #
    ####################################################

    def connect(self):
        host = self.config.get("host")
        port = self.config.get("port")
        user = self.config.get("user")
        _pass = self.config.get("pass")
        sender = self.config.get("sender", user)

        context = ssl.create_default_context()

        if self.config.get("tls", True):
            server = SMTP(host, port)
            server.ehlo()
            server.starttls(context=context)
            server.ehlo()
        else:
            server = SMTP_SSL(host, port, context=context)

        server.login(user, _pass)

        return Mailer(server, sender)


# ----------------------------------------------- #

####################################################
#                      MAILER                      #
####################################################


class Mailer:
    def __init__(self, server: SMTP, sender):
        self.server = server
        self.sender = sender

    def send_custom_email(
        self,
        sender: str = None,
        recipients: list = [],
        subject: str = "Jaseci Email",
        body: tuple = ("", ""),
    ):
        message = MIMEMultipart()
        message["Subject"] = subject
        message["From"] = self.sender if sender is None else sender
        message["To"] = ", ".join(recipients)

        message.attach(MIMEText(body[0], "plain"))
        message.attach(MIMEText(body[1], "html"))
        self.server.sendmail(message["From"], recipients, message.as_string())

    def terminate(self):
        self.server.quit()



from jaseci.utils.test_core import CoreTest, jac_testcase


class ZlibTest(CoreTest):
    fixture_src = __file__

    data_uncompressed = "iVBORw0KGgoAAAANSUhEUgAAAAgAAAAIAQMAAAD+wSzIAAAABlBMVEX///+/v7+jQ3Y5AAAADklEQVQI12P4AIX8EAgALgAD/aNpbtEAAAAASUVORK5CYII="
    data_compressed = "eJzrDPBz5+WS4mJgYOD19HAJAtIcIMzIDCT/HdQ5AaTYAnxCXP///79///7FzmWWQBE+TxfHEI7ryT8YWv8IcDDoMTD/XZyZdxEow+Dp6ueyzimhCQDLCRkd"

    @jac_testcase("zlib.jac", "compress_test", True)
    def test_compress(self, ret):
        print(ret["report"])
        self.assertEqual(
            ret["report"][0],
            self.data_compressed,
        )

    @jac_testcase("zlib.jac", "decompress_test")
    def test_decompress(self, ret):
        print(ret["report"])
        self.assertEqual(
            ret["report"][0],
            self.data_uncompressed,
        )



from jaseci.utils.test_core import CoreTest


class ArchitypeApiTest(CoreTest):
    """Unit tests for Jac Architype APIs"""

    fixture_src = __file__

    def test_architype_register(self):
        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("hello_world.jac")}],
        )

        ret = self.call(
            self.smast,
            [
                "architype_register",
                {
                    "code": self.load_jac("test_architype.jac"),
                    "encoded": False,
                },
            ],
        )

        self.assertEqual(ret["success"], True)
        self.assertEqual(ret["architype"]["name"], "test_architype")

        architype_list_ret = self.call(
            self.smast,
            ["architype_list", {"detailed": True}],
        )

        self.assertEqual(len(architype_list_ret), 7)

    def test_architype_get(self):
        self.call(
            self.smast,
            ["sentinel_register", {"code": self.load_jac("hello_world.jac")}],
        )

        ret = self.call(
            self.smast,
            [
                "architype_register",
                {
                    "code": self.load_jac("test_architype.jac"),
                    "encoded": False,
                },
            ],
        )

        architype_get_ret = self.call(
            self.smast,
            ["architype_get", {"arch": ret["architype"]["jid"]}],
        )

        architype_get_ret_code = self.call(
            self.smast,
            ["architype_get", {"arch": ret["architype"]["jid"], "mode": "code"}],
        )

        self.assertEqual(architype_get_ret["architype"]["name"], "test_architype")
        self.assertEqual(
            "walker test_architype" in architype_get_ret_code["code"], True
        )



"""Built in actions for Jaseci"""
import numpy as np
from operator import itemgetter
import pickle, base64

from jaseci.jsorc.live_actions import jaseci_action


def check_nested_list(lst):
    return all(isinstance(el, list) for el in lst)


@jaseci_action(aliases=["cos_sim"])
def cosine_sim(vec_a: list, vec_b: list):
    """
    Caculate the cosine similarity score of two given vectors
    Param 1 - First vector
    Param 2 - Second vector

    Return - float between 0 and 1
    """
    vec_a_nested = check_nested_list(vec_a)
    vec_b_nested = check_nested_list(vec_b)
    if vec_a_nested or vec_b_nested:
        vec_a_np = np.array(vec_a) if vec_a_nested else np.array([vec_a] * len(vec_b))
        vec_b_np = np.array(vec_b) if vec_b_nested else np.array([vec_b] * len(vec_a))
        sim = np.dot(vec_a_np, vec_b_np.T) / (
            np.linalg.norm(vec_a_np, axis=1)[:, None] * np.linalg.norm(vec_b_np, axis=1)
        )
        return sim.diagonal().tolist()

    result = np.dot(vec_a, vec_b) / (np.linalg.norm(vec_a) * np.linalg.norm(vec_b))
    return float(result.astype(float))


@jaseci_action()
def dot_product(vec_a: list, vec_b: list):
    """
    Caculate the dot product of two given vectors
    Param 1 - First vector
    Param 2 - Second vector

    Return - float between 0 and 1
    """

    return np.inner(vec_a, vec_b).tolist()


@jaseci_action()
def get_centroid(vec_list: list):
    """
    Calculate the centroid of the given list of vectors
    Param 1 - List of vectors

    Return - (centroid vector, cluster tightness)
    """
    centroid = np.mean(vec_list, axis=0)
    tightness = np.mean([cosine_sim(vec, centroid) for vec in vec_list]).astype(float)
    return [centroid.tolist(), tightness]


@jaseci_action()
def softmax(vec_list: list):
    """
    Calculate the centroid of the given list of vectors
    Param 1 - List of vectors

    Return - (centroid vector, cluster tightness)
    """
    e_x = np.exp(vec_list - np.max(vec_list))
    return list(e_x / e_x.sum())


@jaseci_action()
def sort_by_key(data: dict, reverse=False, key_pos=None):
    """
    Sort the given list. Optionally by specific key
    Param 1 - List of items
    Param 2 - if Reverse
    Param 3 (Optional) - Index of the key to be used for sorting
    if param 1 is a list of tuples.

    Deprecated
    """
    if key_pos is not None:
        return sorted(data, key=itemgetter(key_pos), reverse=reverse)
    else:
        return sorted(data, reverse=reverse)


@jaseci_action()
def dim_reduce_fit(data: list, dim=2):
    """
    Dimensionally reduce a list of vectors using incremental PCA
    Param 1 - List of vectors
    Param 2 - Dimension to reduce to

    Return - base64 encoded string of the model
    """
    data_arr = np.array(data)
    ipca = IncrementalPCA(n_components=dim)
    ipca = ipca.fit(data_arr)
    return base64.b64encode(pickle.dumps(ipca)).decode("utf-8")


@jaseci_action()
def dim_reduce_apply(data: list, model: str):
    """
    Dimensionally reduce a list of vectors using incremental PCA
    Param 1 - List of vectors
    Param 2 - base64 encoded string of the model

    Return - List of reduced vectors
    """
    data_arr = np.array(data)
    # reshaping the data if it is a single vector
    if len(data_arr.shape) == 1:
        data_arr = data_arr.reshape(1, -1)
    ipca = pickle.loads(base64.b64decode(model))
    return ipca.transform(data_arr).tolist()


class IncrementalPCA:
    def __init__(self, n_components):
        self.n_components = n_components
        self.mean_ = None
        self.components_ = None
        self.n_samples_seen_ = 0

    def fit(self, X):
        if self.mean_ is None:
            self.mean_ = np.zeros(X.shape[1], dtype=np.float64)
        if self.components_ is None:
            self.components_ = np.zeros(
                (self.n_components, X.shape[1]), dtype=np.float64
            )

        # update mean
        col_mean = np.mean(X, axis=0)
        total_n_samples = self.n_samples_seen_ + X.shape[0]
        self.mean_ = (
            self.n_samples_seen_ * self.mean_ + X.shape[0] * col_mean
        ) / total_n_samples

        # update components
        X_centered = X - self.mean_
        U, S, V = np.linalg.svd(X_centered, full_matrices=False)
        explained_variance = S**2 / (X.shape[0] - 1)
        explained_variance_ratio = explained_variance / np.sum(explained_variance)
        self.components_ = V[: self.n_components]
        self.n_samples_seen_ = total_n_samples

        return self

    def transform(self, X):
        X_centered = X - self.mean_
        return np.dot(X_centered, self.components_.T)



"""
Node class for Jaseci

Each node has an id, name, timestamp and it's set of edges.
First node in list of 'member_node_ids' is designated root node
"""
from collections import OrderedDict
from jaseci.prim.element import Element
from jaseci.prim.obj_mixins import Anchored
from jaseci.prim.edge import Edge
from jaseci.utils.id_list import IdList
from jaseci.utils.utils import logger

import uuid

TO = 0
FROM = 1
BI = 2


class Node(Element, Anchored):
    """Node class for Jaseci"""

    def __init__(self, dimension=0, **kwargs):
        self.edge_ids = IdList(self)
        self.fast_edges = {}  # {name: [[NODEID, DIR, EDGEID, CONTEXT],...]}
        self._fast_edge_ids = IdList(self)
        self.parent_node_ids = IdList(self)
        self.member_node_ids = IdList(self)
        self.dimension = dimension  # Nodes are always hdgd 0
        Element.__init__(self, **kwargs)
        Anchored.__init__(self)

    @property
    def smart_edges(self):
        obj_list = self.smart_edge_list.obj_list()
        for i in obj_list:
            if i.is_fast() and i.jid in self.edge_ids:
                self.edge_ids.remove_obj(i)
                self._fast_edge_ids.remove_obj(i)
                self.smart_add_edge(i)
        return obj_list

    @property
    def smart_edge_list(self):
        if not len(self._fast_edge_ids):
            self.smart_build_fast_edge_ids()
        return self._fast_edge_ids

    def smart_build_fast_edge_ids(self):
        self._fast_edge_ids = IdList(self, in_list=self.edge_ids)
        for k in self.fast_edges.keys():
            for v in self.fast_edges[k]:
                link_order = [v[0], self.jid] if v[1] == FROM else [self.jid, v[0]]
                edge = Edge(
                    m_id=self._m_id, h=self._h, kind="edge", name=k, auto_save=False
                )
                edge.from_node_id = link_order[0]
                edge.to_node_id = link_order[1]
                edge.bidirected = v[1] == BI
                edge.jid = v[2] if len(v) > 2 else uuid.uuid4().urn
                edge.context = v[3] if len(v) > 3 else {}
                edge.save()
                self._fast_edge_ids.add_obj(edge)

    def smart_add_edge(self, obj):
        # make sure fast edges built
        if not len(self._fast_edge_ids):
            self.smart_build_fast_edge_ids()
        # add new edge
        self._fast_edge_ids.add_obj(obj)
        # then store how needed
        if obj.is_fast():
            self.smart_edge_to_fast_edge(obj)
        else:
            self.edge_ids.add_obj(obj)

    def smart_edge_to_fast_edge(self, obj):
        if obj.name not in self.fast_edges:
            self.fast_edges[obj.name] = []
        details = [
            obj.opposing_node(self).jid,
            BI if obj.is_bidirected() else TO if obj.from_node() == self else FROM,
            obj.jid,
            obj.context,
        ]
        self.fast_edges[obj.name].append(details)

    def smart_remove_edge(self, obj):
        if obj.is_fast():
            pluck = None
            for i in self.fast_edges[obj.name]:
                other_node_id = obj.to_node_id if i[1] == TO else obj.from_node_id
                if i[0] == other_node_id:
                    if obj.is_bidirected() or not (
                        i[1] == TO and obj.to_node() == self
                    ):
                        pluck = i
                        break
            self.fast_edges[obj.name].remove(pluck)
            if not len(self.fast_edges[obj.name]):
                del self.fast_edges[obj.name]
        elif obj and obj.jid in self.edge_ids:
            self.edge_ids.remove_obj(obj)
        self.clear_fast_edge_ids()
        self.save()

    def clear_fast_edge_ids(self):
        self._fast_edge_ids = IdList(self)

    def attach(self, node_obj, edge_set=None, as_outbound=True, as_bidirected=False):
        """
        Generalized attach function for attaching nodes with edges
        """
        if edge_set is None:
            edge_set = [
                Edge(
                    m_id=self._m_id,
                    h=self._h,
                    kind="edge",
                    name="generic",
                )
            ]
        link_order = [self, node_obj] if as_outbound else [node_obj, self]
        for e in edge_set:
            if not e.connect(link_order[0], link_order[1]):
                # Node not found error logged in set node function
                return []
            e.set_bidirected(as_bidirected)
        # save and return
        self.save()
        node_obj.save()
        return edge_set

    def attach_outbound(self, node_obj, edge_set=None):
        """
        Creates edge to a node and returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_outbound=True)

    def attach_inbound(self, node_obj, edge_set=None):
        """
        Creates edge from a node and returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_outbound=False)

    def attach_bidirected(self, node_obj, edge_set=None):
        """
        Creates bidirected edge to node returns the edge
        edge_set is the list of edges to be used to make connections
        new edge is created if edge_set is empty
        """
        return self.attach(node_obj, edge_set, as_bidirected=True)

    def detach(
        self,
        node_obj,
        edge_set=None,
        as_outbound=True,
        as_bidirected=False,
        ignore_direction=False,
        silent=True,
    ):
        """
        Generalized detach function for detaching nodes with edges
        """
        if edge_set is None:
            edge_set = self.attached_edges(node_obj)
        link_order = [self, node_obj] if as_outbound else [node_obj, self]
        num_detached = 0
        for e in edge_set:
            # validate edge connection exists
            if not e.connects(
                link_order[0], link_order[1], ignore_direction=ignore_direction
            ):
                if not silent:
                    logger.warning(
                        str(
                            f"{e} does not connect "
                            f"{link_order[0]} to {link_order[1]}"
                        )
                    )
                continue
            if as_bidirected and not ignore_direction and not e.is_bidirected():
                if not silent:
                    logger.warning(str(f"{e} is not a bidirected edge "))
                continue
            # destroy edge
            num_detached += 1
            e.destroy()
        return num_detached

    def detach_outbound(self, node_obj, edge_set=None, silent=True):
        """
        Destroy edges to a node
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_outbound=True)

    def detach_inbound(self, node_obj, edge_set=None, silent=True):
        """
        Destroy edges from a node
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_outbound=False)

    def detach_bidirected(self, node_obj, edge_set=None, silent=True):
        """
        Destroy bidirected edges between nodes
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, as_bidirected=True)

    def detach_edges(self, node_obj, edge_set=None, silent=True):
        """
        Destroy given edges between nodes without checking orientation
        edge_set is the list of edges to be detached (and distroyed)
        all edges are deteached and destroyed if edge_set empty
        returns number of detachments
        """
        return self.detach(node_obj, edge_set, ignore_direction=True)

    def destroy_outbound(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if edge_set is None:
            edge_set = self.outbound_edges(node_obj)
        if self.detach_outbound(node_obj, edge_set):
            node_obj.destroy()

    def destroy_inbound(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if edge_set is None:
            edge_set = self.inbound_edges(node_obj)
        if self.detach_inbound(node_obj, edge_set):
            node_obj.destroy()

    def destroy_bidirected(self, node_obj, edge_set=None):
        """
        Destroys attached node and all relevant edges
        Node and all edges are destroyed if edge_set empty
        """
        if edge_set is None:
            edge_set = self.bidirected_edges(node_obj)
        if self.detach_bidirected(node_obj, edge_set):
            node_obj.destroy()

    def is_attached_out(self, node_obj, edge_set=None):
        """
        Tests whether edges attach to a node
        """
        out_set = self.outbound_edges(node_obj)
        if edge_set is None:
            return len(out_set)
        else:
            for e in edge_set:
                if e not in out_set:
                    return False
        return True

    def is_attached_in(self, node_obj, edge_set=None):
        """
        Tests whether edges attach from a node
        """
        in_set = self.inbound_edges(node_obj)
        if edge_set is None:
            return len(in_set)
        else:
            for e in edge_set:
                if e not in in_set:
                    return False
        return True

    def is_attached_bi(self, node_obj, edge_set=None):
        """
        Tests whether edges attach either to or from a node
        """
        bi_set = self.bidirected_edges(node_obj)
        if edge_set is None:
            return len(bi_set)
        else:
            for e in edge_set:
                if e not in bi_set:
                    return False
        return True

    def outbound_edges(self, node_obj=None):
        """Returns list of all edges out of node"""
        edge_set = []
        for e in self.smart_edges:
            if not e.is_bidirected() and e.connects(self, node_obj):
                edge_set.append(e)
        return edge_set

    def inbound_edges(self, node_obj=None):
        """Returns list of all edges in to node"""
        edge_set = []
        for e in self.smart_edges:
            if not e.is_bidirected() and e.connects(node_obj, self):
                edge_set.append(e)
        return edge_set

    def bidirected_edges(self, node_obj=None):
        """Returns list of all edges between nodes"""
        edge_set = []
        for e in self.smart_edges:
            if e.is_bidirected() and e.connects(self, node_obj):
                edge_set.append(e)
        return edge_set

    def attached_edges(self, node_obj=None, silent=False):
        """
        Returns the edges connecting self to or from a node

        silent is used to indicate whther the edge is intened to be used.
        (effectively turns off error checking when false)
        """
        edge_set = (
            self.outbound_edges(node_obj)
            + self.inbound_edges(node_obj)
            + self.bidirected_edges(node_obj)
        )
        if not silent and edge_set is None:
            logger.error(str(f"No edges found between {self} and {node_obj}"))
        return edge_set

    def outbound_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges out"""
        if edge_set is None:
            edge_set = self.smart_edges
        ret_list = []
        for e in edge_set:
            if not e.is_bidirected() and e.connects(source=self):
                ret_list.append(e.to_node())
        return ret_list

    def inbound_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges in"""
        if edge_set is None:
            edge_set = self.smart_edges
        ret_list = []
        for e in edge_set:
            if not e.is_bidirected() and e.connects(target=self):
                ret_list.append(e.from_node())
        return ret_list

    def bidirected_nodes(self, edge_set=None):
        """Returns list of all nodes connected by edges"""
        if edge_set is None:
            edge_set = self.smart_edges
        ret_list = []
        for e in edge_set:
            if e.is_bidirected():
                ret_list.append(e.opposing_node(self))
        return ret_list

    def attached_nodes(self):
        """Returns list of all nodes connected"""
        edge_set = self.smart_edges
        ret_list = []
        for e in edge_set:
            ret_list.append(e.opposing_node(self))
        return ret_list

    def dimension_matches(self, node_obj, silent=True):
        """Test if dimension matches another node"""
        matches = self.dimension == node_obj.dimension
        if not matches and not silent:
            logger.error(
                str(
                    "'{}' cant connect to '{}' - dim mismatch {}->{}".format(
                        self, node_obj, self.dimension, node_obj.dimension
                    )
                )
            )
        return matches

    def make_member_of(self, node_obj):
        """
        Adds node to higher dimension node and does relevant checks
        """
        # check if valid inclusion
        if node_obj.dimension != self.dimension + 1:
            logger.error(
                str(
                    "'{}' cant be member to '{}' - dimension mismatch {}->{}".format(
                        self, node_obj, self.dimension, node_obj.dimension
                    )
                )
            )
        # adds self to hdgd and hdgd to list of owners
        else:
            node_obj.member_node_ids.add_obj(self)
            self.parent_node_ids.add_obj(node_obj)

    def make_owner_of(self, node_obj):
        """
        Adds node to lower dimension node and does relevant checks
        """
        node_obj.make_member_of(self)

    def leave_memebership_of(self, node_obj):
        """Remove node from higher dimension node"""
        node_obj.member_node_ids.remove_obj(self)
        self.parent_node_ids.remove_obj(node_obj)

    def disown(self, node_obj):
        """Remove node from higher dimension node"""
        node_obj.leave_membership_of(self)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        for i in self.smart_edges:
            i.destroy()
        super().destroy()

    def dot_str(self, node_map=None, detailed=False):
        """
        DOT representation
        """

        def handle_str(str):
            return str[:32].replace('"', '\\"')

        if node_map is None:
            nid = f"{uuid.UUID(self.jid).hex}"
        else:
            nid = f"{node_map.index(self.jid)}"

        dstr = f'"n{nid}" [ '

        if detailed:
            dstr += f'id="{uuid.UUID(self.jid).hex}", '

        dstr += f'label="n{nid}:{self.name}" '

        node_dict = self.context
        for i in self.private_values():
            node_dict.pop(i)

        if node_dict and detailed:
            for k, v in node_dict.items():
                if not isinstance(v, str) or v == "":
                    continue
                dstr += f', {k}="{handle_str(v)}"'

        dstr += " ]"

        return dstr + "\n"

    def get_all_architypes(self, depth: int = 0):
        """
        Returns all reachable architypes
        """

        childs = {self.jid: self}

        edges = OrderedDict()
        nodes = OrderedDict(childs)

        depth -= 1

        while len(childs) and depth != 0:
            new_childs = OrderedDict()

            for node in childs.values():
                for edge in node.attached_edges():
                    if not (edge.jid in edges):
                        n_node = False
                        to_node = edge.to_node()

                        if to_node == node:
                            n_node = edge.from_node()
                        else:
                            n_node = to_node

                        if not (n_node.jid in nodes):
                            edges.update({edge.jid: edge})
                            new_childs.update({n_node.jid: n_node})
                        else:
                            edges.update({edge.jid: edge})

            childs = new_childs
            nodes.update(childs)
            depth -= 1

        return nodes, edges

    def traversing_dot_str(self, detailed=False, depth: int = 0):
        """
        DOT representation for graph.
        NOTE: This is different from the dot_str method for node intentionally
        because graph inherits node.
        """
        nodes, edges = self.get_all_architypes(depth)

        # Construct the graph string
        dstr = ""
        dstr += f"strict digraph {self.name} {{\n"
        for n in nodes.values():
            dstr += f"    {n.dot_str(list(nodes.keys()), detailed)}"
        for e in edges.values():
            dstr += f"    {e.dot_str(list(nodes.keys()), list(edges.keys()), detailed)}"
        dstr += "}"
        return dstr



"""
Walker class for Jaseci

Each walker has an id, name, timestamp and it's parent sentinel.
TODO: Perhaps  I should have walker state (context ids) in mem only with
default hooks to save db read/writes
"""

from jaseci.utils.utils import (
    logger,
    perf_test_start,
    perf_test_stop,
    perf_test_to_b64,
    exc_stack_as_str_list,
)
from jaseci.prim.element import Element
from jaseci.prim.obj_mixins import Anchored
from jaseci.utils.id_list import IdList
from jaseci.jac.interpreter.walker_interp import WalkerInterp
import uuid
import hashlib

from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.task_svc import TaskService


class Walker(Element, WalkerInterp, Anchored):
    """Walker class for Jaseci"""

    def __init__(self, is_async=False, **kwargs):
        self.yielded = False
        self.namespaces = []
        self.profile = {}
        # Process state
        self.current_node_id = None
        self.next_node_ids = IdList(self)
        self.destroy_node_ids = IdList(self)
        self.current_step = 0
        self.in_entry_exit = False
        self.step_limit = 10000
        self.is_async = is_async
        self._to_await = False
        Element.__init__(self, **kwargs)
        WalkerInterp.__init__(self)
        Anchored.__init__(self)

    @property
    def current_node(self):
        if not self.current_node_id:
            return None
        elif not self._h.has_obj(self.current_node_id):
            self.current_node_id = None
            return None
        else:
            return self._h.get_obj(self._m_id, self.current_node_id)

    @current_node.setter
    def current_node(self, obj):
        if obj:
            self.current_node_id = obj.jid
        else:
            self.current_node_id = None

    def namespace_keys(self):
        """Return list of md5 keys for namespaces"""
        ret = {}
        for i in self.namespaces:
            ret[i] = hashlib.md5((self._m_id + i).encode()).hexdigest()
        return ret

    def for_queue(self):
        return self.is_async and not (self._to_await)

    def step(self):
        """
        Take single step through program
        if no ast provided, will be generated from code
        """
        if not self.next_node_ids:
            logger.debug(str(f"Walker {self.name} is disengaged"))
            return False
        if self.current_step > self.step_limit:
            logger.error(
                str(
                    f"Walker {self.name} walked too many steps "
                    + f"- {self.step_limit}"
                )
            )
            return False

        self.current_node = self.next_node_ids.pop_first_obj()
        self.run_walker(jac_ast=self.get_architype().get_jac_ast())
        if self.current_step < 200:
            self.log_history("visited", self.current_node.id)
        self.current_step += 1
        self.profile["steps"] = self.current_step
        if self._stopped == "skip":
            self._stopped = None
        if self.next_node_ids:
            logger.debug(
                str(
                    f"Step complete, Walker {self.name} next node: "
                    + f"- {self.next_node_ids.first_obj()}"
                )
            )
            return self.next_node_ids.first_obj()
        else:
            logger.debug(
                str(
                    f"Final step of walker {self.name} complete"
                    + f"- disengaging from {self.current_node}"
                )
            )
            for i in self.destroy_node_ids.obj_list():
                if i.jid == self.current_node_id:
                    self.current_node_id = None
                i.destroy()
                self.destroy_node_ids.remove_obj(i)
            return True

    def prime(self, start_node, prime_ctx=None, request_ctx=None):
        """Place walker on node and get ready to step step"""
        if not self.yielded:
            self.clear_state()
        if not self.yielded or not len(self.next_node_ids):  # modus ponens
            self.next_node_ids.add_obj(start_node, push_front=True)
        if prime_ctx:
            for i in prime_ctx.keys():
                self.context[str(i)] = prime_ctx[i]
        self.request_context = request_ctx
        self.profile["steps"] = self.current_step
        logger.debug(str(f"Walker {self.name} primed - {start_node}"))

    def run(self, start_node=None, prime_ctx=None, request_ctx=None, profiling=False):
        """Executes Walker to completion"""
        if self.for_queue() and JsOrc.svc("task").is_running():
            start_node = (
                start_node
                if not (start_node is None)
                else (
                    self.next_node_ids.pop_first_obj() if self.next_node_ids else None
                )
            )

            self._h.commit_all_cache_sync()

            return {
                "is_queued": True,
                "result": JsOrc.svc("task", TaskService).add_queue(
                    self,
                    start_node,
                    prime_ctx or self.context,
                    request_ctx or self.request_context,
                    profiling,
                ),
            }

        if profiling:
            pr = perf_test_start()

        if start_node and (not self.yielded or not len(self.next_node_ids)):
            self.prime(start_node, prime_ctx, request_ctx)
        elif prime_ctx:
            for i in prime_ctx.keys():
                self.context[str(i)] = prime_ctx[i]

        report_ret = {"success": True}
        WalkerInterp.reset(self)
        self.yielded = False

        try:
            while self.step() and not self.yielded:
                pass
        except Exception as e:
            self.rt_error(f"Internal Exception: {e}", self._cur_jac_ast)
            report_ret["stack_trace"] = exc_stack_as_str_list()

        self.save()

        if not self.report:
            logger.debug(str(f"Walker {self.name} did not have anything to report"))
        report_ret["report"] = self.report
        report_ret["final_node"] = self.current_node_id
        report_ret["yielded"] = self.yielded

        if self.report_status:
            report_ret["status_code"] = self.report_status
        if self.report_custom:
            report_ret["report_custom"] = self.report_custom
        if len(self.runtime_errors):
            report_ret["errors"] = self.runtime_errors
            report_ret["success"] = False
        if profiling:
            self.profile["perf"] = perf_test_stop(pr)
            self.profile["graph"] = perf_test_to_b64(pr)
            report_ret["profile"] = self.profile

        if self.for_queue():
            return {"is_queued": False, "result": report_ret}

        return report_ret

    def yield_walk(self):
        """Instructs walker to yield (stop walking and keep state)"""
        self.yielded = True

    def log_history(self, name, value):
        """Helper function for logging history of walker's activities"""
        if isinstance(value, Element):
            value = {"type": value.j_type, "id": value.jid}
        if isinstance(value, uuid.UUID):
            value = value.urn
        if name in self.profile.keys():
            self.profile[name].append(value)
        else:
            self.profile[name] = [value]

    def clear_state(self):
        """Clears walker state after report"""
        self.yielded = False
        self.profile = {}
        self.current_step = 0
        self.next_node_ids.remove_all()
        self.ignore_node_ids.remove_all()
        self.destroy_node_ids.remove_all()
        self.current_node = None
        self.context = {}
        WalkerInterp.reset(self)

    def destroy(self):
        """
        Destroys self from memory and persistent storage
        """
        if not self.for_queue() or not JsOrc.svc("task").is_running():
            WalkerInterp.destroy(self)
            super().destroy()

    def register_yield_or_destroy(self, yield_ids):
        """Helper for auto destroying walkers"""
        if not self.yielded:
            if self.jid in yield_ids:
                yield_ids.remove_obj(self)
            self.destroy()
        else:
            yield_ids.add_obj(self, silent=True)

    def save(self):
        """
        Write self through hook to persistent storage
        """
        self._h.save_obj(caller_id=self._m_id, item=self, all_caches=self.is_async)



from jaseci.jac.ir.passes import (
    ParseTreePrunePass,
    PrinterPass,
    StatsPass,
    CodeGenPass,
    AstPrunePass,
)
from jaseci.jac.ir.ast import Ast


def multi_pass_optimizer(jac_ast: Ast, opt_level: int):
    ParseTreePrunePass(ir=jac_ast).run() if opt_level > 0 else None
    CodeGenPass(ir=jac_ast).run() if opt_level > 2 else None
    AstPrunePass(ir=jac_ast).run() if opt_level > 2 else None


def debug_pass(
    jac_ast: Ast,
    print_walk: bool = False,
    stats_out: bool = False,
):
    PrinterPass(jac_ast).run() if print_walk else None
    StatsPass(jac_ast).run() if stats_out else None



"""Built in actions for Jaseci"""
from jaseci.jsorc.jsorc import JsOrc
from jaseci.jsorc.live_actions import jaseci_action
from jaseci.extens.svc.task_svc import TaskService


@jaseci_action()
def get_result(task_id, wait, meta):
    """
    Get task result by task_id
    """

    return JsOrc.svc("task").poke(TaskService).get_by_task_id(task_id, wait)



from jaseci.jsorc.live_actions import jaseci_action
from jaseci.jsorc.jsorc import JsOrc
from jaseci.extens.svc.elastic_svc import ElasticService, Elastic


def elastic():
    return JsOrc.svc("elastic", ElasticService).poke(Elastic)


@jaseci_action()
def _post(url: str, json: dict = {}):
    return elastic()._post(url, json)


@jaseci_action()
def _get(url: str, json: dict = None):
    return elastic()._get(url, json)


@jaseci_action()
def post(url: str, body: dict, index: str = "", suffix: str = ""):
    return elastic().post(url, body, index, suffix)


@jaseci_action()
def post_act(url: str, body: dict, suffix: str = ""):
    return elastic().post_act(url, body, suffix)


@jaseci_action()
def get(url: str, body: dict, index: str = "", suffix: str = ""):
    return elastic().get(url, body, index)


@jaseci_action()
def get_act(url: str, body: dict, suffix: str = ""):
    return elastic().get_act(url, body, suffix)


@jaseci_action()
def doc(log: dict, query: str = "", index: str = "", suffix: str = ""):
    return elastic().doc(log, query, index, suffix)


@jaseci_action()
def doc_activity(log: dict, query: str = "", suffix: str = ""):
    return elastic().doc_activity(log, query, suffix)


@jaseci_action()
def search(body: dict, query: str = "", index: str = "", suffix: str = ""):
    return elastic().search(body, query, index, suffix)


@jaseci_action()
def search_activity(body: dict, query: str = "", suffix: str = ""):
    return elastic().search_activity(body, query, suffix)


@jaseci_action()
def mapping(query: str = "", index: str = "", suffix: str = ""):
    return elastic().mapping(query, index, suffix)


@jaseci_action()
def mapping_activity(query: str = "", suffix: str = ""):
    return elastic().mapping_activity(query, suffix)


@jaseci_action()
def refresh(index: str = "", suffix: str = ""):
    return elastic().refresh(index, suffix)


@jaseci_action()
def refresh_activity(suffix: str = ""):
    return elastic().refresh_activity(suffix)


@jaseci_action()
def aliases(query: str = "pretty=true"):
    return elastic().aliases(query)


@jaseci_action()
def reindex(body: dict, query: str = "pretty"):
    return elastic().reindex(body, query)



class IrPass:
    def __init__(self, ir=None):
        self.ir = ir

    def before_pass(self):
        pass

    def after_pass(self):
        pass

    def enter_node(self, node):
        pass

    def exit_node(self, node):
        pass

    def run(self):
        self.before_pass()
        self.traverse()
        self.after_pass()
        return self

    def traverse(self, node=None):
        if node is None:
            node = self.ir
        self.enter_node(node)
        for i in node.kid:
            self.traverse(i)
        self.exit_node(node)
